/* sh-stub.c -- debugging stub for the Hitachi-SH. 

 NOTE!! This code has to be compiled with optimization, otherwise the 
 function inlining which generates the exception handlers won't work.
 
*/

/*   This is originally based on an m68k software stub written by Glenn
     Engel at HP, but has changed quite a bit. 

     Modifications for the SH by Ben Lee and Steve Chamberlain

*/

/****************************************************************************

		THIS SOFTWARE IS NOT COPYRIGHTED

   HP offers the following for use in the public domain.  HP makes no
   warranty with regard to the software or it's performance and the
   user accepts the software "AS IS" with all faults.

   HP DISCLAIMS ANY WARRANTIES, EXPRESS OR IMPLIED, WITH REGARD
   TO THIS SOFTWARE INCLUDING BUT NOT LIMITED TO THE WARRANTIES
   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

****************************************************************************/


/* Remote communication protocol.

   A debug packet whose contents are <data>
   is encapsulated for transmission in the form:

	$ <data> # CSUM1 CSUM2

	<data> must be ASCII alphanumeric and cannot include characters
	'$' or '#'.  If <data> starts with two characters followed by
	':', then the existing stubs interpret this as a sequence number.

	CSUM1 and CSUM2 are ascii hex representation of an 8-bit 
	checksum of <data>, the most significant nibble is sent first.
	the hex digits 0-9,a-f are used.

   Receiver responds with:

	+	- if CSUM is correct and ready for next packet
	-	- if CSUM is incorrect

   <data> is as follows:
   All values are encoded in ascii hex digits.

	Request		Packet

	read registers  g
	reply		XX....X		Each byte of register data
					is described by two hex digits.
					Registers are in the internal order
					for GDB, and the bytes in a register
					are in the same order the machine uses.
			or ENN		for an error.

	write regs	GXX..XX		Each byte of register data
					is described by two hex digits.
	reply		OK		for success
			ENN		for an error

        write reg	Pn...=r...	Write register n... with value r...,
					which contains two hex digits for each
					byte in the register (target byte
					order).
	reply		OK		for success
			ENN		for an error
	(not supported by all stubs).

	read mem	mAA..AA,LLLL	AA..AA is address, LLLL is length.
	reply		XX..XX		XX..XX is mem contents
					Can be fewer bytes than requested
					if able to read only part of the data.
			or ENN		NN is errno

	write mem	MAA..AA,LLLL:XX..XX
					AA..AA is address,
					LLLL is number of bytes,
					XX..XX is data
	reply		OK		for success
			ENN		for an error (this includes the case
					where only part of the data was
					written).

	cont		cAA..AA		AA..AA is address to resume
					If AA..AA is omitted,
					resume at same address.

	step		sAA..AA		AA..AA is address to resume
					If AA..AA is omitted,
					resume at same address.

	last signal     ?               Reply the current reason for stopping.
                                        This is the same reply as is generated
					for step or cont : SAA where AA is the
					signal number.

	There is no immediate reply to step or cont.
	The reply comes when the machine stops.
	It is		SAA		AA is the "signal number"

	or...		TAAn...:r...;n:r...;n...:r...;
					AA = signal number
					n... = register number
					r... = register contents
	or...		WAA		The process exited, and AA is
					the exit status.  This is only
					applicable for certains sorts of
					targets.
	kill request	k

	toggle debug	d		toggle debug flag (see 386 & 68k stubs)
	reset		r		reset -- see sparc stub.
	reserved	<other>		On other requests, the stub should
					ignore the request and send an empty
					response ($#<checksum>).  This way
					we can extend the protocol and GDB
					can tell whether the stub it is
					talking to uses the old or the new.
	search		tAA:PP,MM	Search backwards starting at address
					AA for a match with pattern PP and
					mask MM.  PP and MM are 4 bytes.
					Not supported by all stubs.

	general query	qXXXX		Request info about XXXX.
	general set	QXXXX=yyyy	Set value of XXXX to yyyy.
	query sect offs	qOffsets	Get section offsets.  Reply is
					Text=xxx;Data=yyy;Bss=zzz
	console output	Otext		Send text to stdout.  Only comes from
					remote target.

	Responses can be run-length encoded to save space.  A '*' means that
	the next character is an ASCII encoding giving a repeat count which
	stands for that many repititions of the character preceding the '*'.
	The encoding is n+29, yielding a printable character where n >=3 
	(which is where rle starts to win).  Don't use an n > 126. 

	So 
	"0* " means the same as "0000".  */

/*
 *  @(#) $Id: sh-stub.7145.c,v 1.1.1.3 2004/10/19 12:36:50 hori Exp $
 */

#include "sh-sci.7145.h"
#include "sh-stub.h"
#include "mini_lib.h"

/*
 *  쥸ѷ
 */
typedef unsigned char   byte;
typedef volatile byte   IOREG;
typedef unsigned short  word;
typedef volatile word   HIOREG;
typedef volatile int    LIOREG;

/*
 * Хơȥȥ
 * ԥե󥯥󥳥ȥ
 */
#define BCR1	((HIOREG *)0xffff8620)
#define BCR2	((HIOREG *)0xffff8622)
#define WCR1	((HIOREG *)0xffff8624)
#define PACRH	((HIOREG *)0xffff8388)
#define PACRL1	((HIOREG *)0xffff838c)
#define PACRL2	((HIOREG *)0xffff838e)
#define PBCR1	((HIOREG *)0xffff8398)
#define PBCR2	((HIOREG *)0xffff839a)
#define PCCR	((HIOREG *)0xffff839c)
#define PDCRH1	((HIOREG *)0xffff83a8)
#define PDCRH2	((HIOREG *)0xffff83aa)
#define PDCRL1	((HIOREG *)0xffff83ac)
#define PDCRL2	((HIOREG *)0xffff83ae)
#define PECRL1	((HIOREG *)0xffff83b8)
#define PECRL2	((HIOREG *)0xffff83ba)
#define PAIORL	((HIOREG *)0xffff8386)
#define PBIOR	((HIOREG *)0xffff8394)
#define PEIORL	((HIOREG *)0xffff83b4)
#define PEDRL	((HIOREG *)0xffff83b0)

/*
 * ߥȥ쥸
 */
#define IPRA    ((HIOREG *)0xffff8348)
#define IPRB    ((HIOREG *)0xffff834a)
#define IPRC    ((HIOREG *)0xffff834c)
#define IPRD    ((HIOREG *)0xffff834e)
#define IPRE    ((HIOREG *)0xffff8350)
#define IPRF    ((HIOREG *)0xffff8352)
#define IPRG    ((HIOREG *)0xffff8354)
#define IPRH    ((HIOREG *)0xffff8356)



#define COND_BR_MASK	0xff00
#define UCOND_DBR_MASK	0xe000
#define UCOND_RBR_MASK	0xf0df
#define TRAPA_MASK		0xff00

#define COND_DISP		0x00ff
#define UCOND_DISP		0x0fff
#define UCOND_REG		0x0f00

#define BF_INSTR		0x8b00
#define BT_INSTR		0x8900
#define BFS_INSTR		0x8f00
#define BTS_INSTR		0x8d00
#define BRA_INSTR		0xa000
#define BRAF_INSTR		0x0023
#define BSRF_INSTR		0x0003
#define BSR_INSTR		0xb000
#define JMP_INSTR		0x402b
#define JSR_INSTR		0x400b
#define RTS_INSTR		0x000b
#define RTE_INSTR		0x002b
#define TRAPA_INSTR		0xc300

#define SSTEP_INSTR		0xc3ff

#define T_BIT_MASK		0x0001
/*
 * BUFMAX defines the maximum number of characters in inbound/outbound
 * buffers at least NUMREGBYTES*2 are needed for register packets
 */
#define BUFMAX 1024

/*
 * Number of bytes for registers
 */
#define NUMREGBYTES 92		/* ꥸʥǤǤ112*/

/*
 * typedef
 */
typedef void (*Function) ();

/*
 * Forward declarations
 */

static int hex (char);
static char *mem2hex (char *, char *, int);
static char *hex2mem (char *, char *, int);
static int hexToInt (char **, int *);
static void getpacket (char *);
static void putpacket (char *);
static void handle_buserror (void);
static int computeSignal (int exceptionVector);
static void handle_exception (int exceptionVector);
static void handle_bios_call (void);

/* These are in the file but in asm statements so the compiler can't see them */
void catch_exception_4 (void);
void catch_exception_6 (void);
void catch_exception_9 (void);
void catch_exception_10 (void);
void catch_exception_11 (void);
void catch_exception_32 (void);
void catch_exception_33 (void);
void catch_exception_255 (void);




void breakpoint (void);

#if 0
int init_stack[init_stack_size] __attribute__ ((section (".stack"))) = {0};
int stub_stack[stub_stack_size] __attribute__ ((section (".stack"))) = {0};
#else
int init_stack[init_stack_size] __attribute__ ((section (".stack")));
int stub_stack[stub_stack_size] __attribute__ ((section (".stack")));
#endif

#define vector_size	256

/*  ٥ơ֥  */
static Function vectable[vector_size];

void INIT ();
void BINIT ();

void Sci_Initialize();

/* When you link take care that this is at address 0 -
   or wherever your vbr points */

#define CPU_BUS_ERROR_VEC  9
#define DMA_BUS_ERROR_VEC 10
#define NMI_VEC           11
#define INVALID_INSN_VEC   4
#define INVALID_SLOT_VEC   6
#define TRAP_VEC          32
#define BIOS_CALL_TRAP    33
#define USER_VEC         255

/*  ͽѤߤΥ٥ֹΥꥹ  */
static int reservation_list[] = {5,7,8,		/*  ƥͽ  */
	/*  ꥢ  */
	/*  detouch塢ץꥱǤ⥷ꥢݡȤ */
	/*  ѤϡΥꥹȤʤƤɤ  */
	SCI_VEC,				
	NMI_VEC, TRAP_VEC, BIOS_CALL_TRAP};

#define ARRAY_SIZE(array)	(sizeof(array)/sizeof(array[0]))

static int is_vector_ok (int n);
static int set_handler(int n, Function func);

char in_nmi;   /* Set when handling an NMI, so we don't reenter */
int dofault;  /* Non zero, bus errors will raise exception */

int *stub_sp;

/* debug > 0 prints ill-formed commands in valid packets & checksum errors */
int remote_debug;

/* jump buffer used for setjmp/longjmp */
jmp_buf remcomEnv;

enum regnames
  {
    R0, R1, R2, R3, R4, R5, R6, R7,
    R8, R9, R10, R11, R12, R13, R14,
    R15, PC, PR, GBR, VBR, MACH, MACL, SR,
    TICKS, STALLS, CYCLES, INSTS, PLR
  };

typedef struct
  {
    short *memAddr;
    short oldInstr;
  }
stepData;

int registers[NUMREGBYTES / 4];
stepData instrBuffer;
char stepped;
static const char hexchars[] = "0123456789abcdef";
char remcomInBuffer[BUFMAX];
char remcomOutBuffer[BUFMAX];

int timercheck=0;
int scicheck=0;

/*
 *  gdb stubˤ
 */

static __inline__ int
stub_putc(int c)
{
  __asm__ volatile("mov #0x00,r0; mov %0,r4; trapa #0x21"
               : /* no output */
               : "r"(c)
               : "r0","r4");
  return(c);
}

void message(char *p)
{
	for(; *p; p++) {
		stub_putc(*p);
	}
}

char highhex(int  x)
{
  return hexchars[(x >> 4) & 0xf];
}

char lowhex(int  x)
{
  return hexchars[x & 0xf];
}

/*
 * Assembly macros
 */

#define BREAKPOINT()   asm("trapa	#0x20"::);


/*
 * Routines to handle hex data
 */

static int
hex (char ch)
{
  if ((ch >= 'a') && (ch <= 'f'))
    return (ch - 'a' + 10);
  if ((ch >= '0') && (ch <= '9'))
    return (ch - '0');
  if ((ch >= 'A') && (ch <= 'F'))
    return (ch - 'A' + 10);
  return (-1);
}

/* convert the memory, pointed to by mem into hex, placing result in buf */
/* return a pointer to the last char put in buf (null) */
static char *
mem2hex (char *mem, char *buf, int count)
{
  int i;
  int ch;
  for (i = 0; i < count; i++)
    {
      ch = *mem++;
      *buf++ = highhex (ch);
      *buf++ = lowhex (ch);
    }
  *buf = 0;
  return (buf);
}

/* convert the hex array pointed to by buf into binary, to be placed in mem */
/* return a pointer to the character after the last byte written */

static char *
hex2mem (char *buf, char *mem, int count)
{
  int i;
  unsigned char ch;
  for (i = 0; i < count; i++)
    {
      ch = hex (*buf++) << 4;
      ch = ch + hex (*buf++);
      *mem++ = ch;
    }
  return (mem);
}

/**********************************************/
/* WHILE WE FIND NICE HEX CHARS, BUILD AN INT */
/* RETURN NUMBER OF CHARS PROCESSED           */
/**********************************************/
static int
hexToInt (char **ptr, int *intValue)
{
  int numChars = 0;
  int hexValue;

  *intValue = 0;

  while (**ptr)
    {
      hexValue = hex (**ptr);
      if (hexValue >= 0)
	{
	  *intValue = (*intValue << 4) | hexValue;
	  numChars++;
	}
      else
	break;

      (*ptr)++;
    }

  return (numChars);
}

/*
 * Routines to get and put packets
 */

/* scan for the sequence $<data>#<checksum>     */

static
void
getpacket (char *buffer)
{
  unsigned char checksum;
  unsigned char xmitcsum;
  int i;
  int count;
  char ch;
  do
    {
      /* wait around for the start character, ignore all other characters */
      while ((ch = getDebugChar ()) != '$');
      checksum = 0;
      xmitcsum = -1;

      count = 0;

      /* now, read until a # or end of buffer is found */
      while (count < BUFMAX)
	{
	  ch = getDebugChar ();
	  if (ch == '#')
	    break;
	  checksum = checksum + ch;
	  buffer[count] = ch;
	  count = count + 1;
	}
      buffer[count] = 0;

      if (ch == '#')
	{
	  xmitcsum = hex (getDebugChar ()) << 4;
	  xmitcsum += hex (getDebugChar ());
	  if (checksum != xmitcsum)
	    putDebugChar ('-');	/* failed checksum */
	  else
	    {
	      putDebugChar ('+');	/* successful transfer */
	      /* if a sequence char is present, reply the sequence ID */
	      if (buffer[2] == ':')
		{
		  putDebugChar (buffer[0]);
		  putDebugChar (buffer[1]);
		  /* remove sequence chars from buffer */
		  count = strlen (buffer);
		  for (i = 3; i <= count; i++)
		    buffer[i - 3] = buffer[i];
		}
	    }
	}
    }
  while (checksum != xmitcsum);

}


/* send the packet in buffer.  The host get's one chance to read it.
   This routine does not wait for a positive acknowledge.  */

static void
putpacket (register char *buffer)
{
  register  int checksum;
//  register  int count;

  /*  $<packet info>#<checksum>. */
  do
    {
      char *src = buffer;
      putDebugChar ('$');
      checksum = 0;

      while (*src)
	{
	  int runlen;

	  /* Do run length encoding */
	  for (runlen = 0; runlen < 100; runlen ++) 
	    {
	      if (src[0] != src[runlen]) 
		{
		  if (runlen > 3) 
		    {
		      int encode;
		      /* Got a useful amount */
		      putDebugChar (*src);
		      checksum += *src;
		      putDebugChar ('*');
		      checksum += '*';
		      checksum += (encode = runlen + ' ' - 4);
		      putDebugChar (encode);
		      src += runlen;
		    }
		  else
		    {
		      putDebugChar (*src);
		      checksum += *src;
		      src++;
		    }
		  break;
		}
	    }
	}

      putDebugChar ('#');
      putDebugChar (highhex(checksum));
      putDebugChar (lowhex(checksum));
    }
  while  (getDebugChar() != '+');

}


/* a bus error has occurred, perform a longjmp
   to return execution and allow handling of the error */

void
handle_buserror (void)
{
  longjmp (remcomEnv, 1);
}

/*
 * this function takes the SH-1 exception number and attempts to
 * translate this number into a unix compatible signal value
 */
static int
computeSignal (int exceptionVector)
{
  int sigval;
  switch (exceptionVector)
    {
    case INVALID_INSN_VEC:
      sigval = 4;
      break;			
    case INVALID_SLOT_VEC:
      sigval = 4;
      break;			
    case CPU_BUS_ERROR_VEC:
      sigval = 10;
      break;			
    case DMA_BUS_ERROR_VEC:
      sigval = 10;
      break;	
    case NMI_VEC:
      sigval = 2;
      break;	

    case TRAP_VEC:
    case USER_VEC:
      sigval = 5;
      break;

    default:
      sigval = 7;		/* "software generated"*/
      break;
    }
  return (sigval);
}

void
doSStep (void)
{
  short *instrMem;
  int displacement;
  int reg;
  unsigned short opcode;

  instrMem = (short *) registers[PC];

  opcode = *instrMem;
  stepped = 1;

  if ((opcode & COND_BR_MASK) == BT_INSTR) /* BT */
    {
      if (registers[SR] & T_BIT_MASK)
	{
	/*  8ӥåȡǥץ졼ȡĥ2  */
	  displacement = opcode & COND_DISP;
	  if (displacement & 0x80)
	    displacement |= 0xffffff00;
	  displacement <<= 1;
		  /*
		   * Remember PC points to second instr.
		   * after PC of branch ... so add 4
		   */
	  instrMem = (short *) (registers[PC] + displacement + 4);
	}
      else
	instrMem += 1;
    }
  else if ((opcode & COND_BR_MASK) == BF_INSTR) /* BF */
    {
      if (registers[SR] & T_BIT_MASK)
	instrMem += 1;
      else
	{
	/*  8ӥåȡǥץ졼ȡĥ2  */
	  displacement = opcode & COND_DISP;
	  if (displacement & 0x80)
	    displacement |= 0xffffff00;
	  displacement <<=1;
		  /*
		   * Remember PC points to second instr.
		   * after PC of branch ... so add 4
		   */
	  instrMem = (short *) (registers[PC] + displacement + 4);
	}
    }
  else if ((opcode & COND_BR_MASK) == BTS_INSTR) /* BTS */
    {
      if (registers[SR] & T_BIT_MASK)
	{
	  displacement = (opcode & COND_DISP);
	  if (displacement & 0x80)   /* ĥ */
	    displacement |= 0xffffff00;
      displacement <<=1;     /*  2  */
	  /*
		   * Remember PC points to second instr.
		   * after PC of branch ... so add 4
		   */
	  instrMem = (short *) (registers[PC] + displacement + 4);
	}
      else
	instrMem += 2;		/* We should not place trapa at the slot */
    }
  else if ((opcode & COND_BR_MASK) == BFS_INSTR) /* BFS */
    {
      if (registers[SR] & T_BIT_MASK)
	instrMem += 2; 		/* We should not place trapa at the slot */
      else
	{
	  displacement = (opcode & COND_DISP);
	  if (displacement & 0x80)    /* ĥ */
	    displacement |= 0xffffff00;
      displacement <<=1;     /*  2  */
	  /*
		   * Remember PC points to second instr.
		   * after PC of branch ... so add 4
		   */
	  instrMem = (short *) (registers[PC] + displacement + 4);
	}
    }
  else if ((opcode & UCOND_DBR_MASK) == BRA_INSTR) /* BRA/BSR */
    {
      /*  12ӥåȡǥץ졼ȡĥ2  */
      displacement = opcode & UCOND_DISP;
      if (displacement & 0x0800)
	displacement |= 0xfffff000;
      displacement <<= 1;	

	  /*
	   * Remember PC points to second instr.
	   * after PC of branch ... so add 4
	   */
      instrMem = (short *) (registers[PC] + displacement + 4);
    }
  else if ((opcode & UCOND_RBR_MASK) == JSR_INSTR) /* JMP/JSR */
    {
      reg = (char) ((opcode & UCOND_REG) >> 8);

      instrMem = (short *) registers[reg];
    }
  else if ((opcode & UCOND_RBR_MASK) == BSRF_INSTR) /* BRAF/BSRF */
    {
      reg = (char) ((opcode & UCOND_REG) >> 8);

      instrMem = (short *) (registers[reg] + (registers[PC]&0xfffffffc) + 4);
    }
  else if (opcode == RTS_INSTR)
    instrMem = (short *) registers[PR];
  else if (opcode == RTE_INSTR)
    instrMem = (short *) registers[15];
  else if ((opcode & TRAPA_MASK) == TRAPA_INSTR) {
    int offset;
    /*  ٥ơ֥뤫ʬɤ߼  */
    offset = opcode & ~TRAPA_MASK;
    if (!is_vector_ok(offset)) {
    	/*  ʬ褬stubξ硢  */
    	/*  stubΤROMˤΤǥƥå׼¹ԤǤʤ  */
    	message("This trap vector can\'t be stepped.\n");
    	message("(reserved for gdb stub on ROM.)\n");
    	message("GDB stub is terminated.\n");
    	while(1);
    }
    offset <<= 2;
    instrMem = *(short **)(registers[VBR] + offset);
  } else
    instrMem += 1;

  instrBuffer.memAddr = instrMem;
  instrBuffer.oldInstr = *instrMem;
  *instrMem = SSTEP_INSTR;
}


/* Undo the effect of a previous doSStep.  If we single stepped,
   restore the old instruction. */

void
undoSStep (void)
{
  if (stepped)
    {  short *instrMem;
      instrMem = instrBuffer.memAddr;
      *instrMem = instrBuffer.oldInstr;
    }
  stepped = 0;
}

/*
This function does all exception handling.  It only does two things -
it figures out why it was called and tells gdb, and then it reacts
to gdb's requests.

When in the monitor mode we talk a human on the serial line rather than gdb.

*/


void
gdb_handle_exception (int exceptionVector)
{
  int sigval;
  int addr, length;
  char *ptr;

  /* reply to host that an exception has occurred */
  sigval = computeSignal (exceptionVector);
  remcomOutBuffer[0] = 'S';
  remcomOutBuffer[1] = highhex(sigval);
  remcomOutBuffer[2] = lowhex (sigval);
  remcomOutBuffer[3] = 0;


  putpacket (remcomOutBuffer);

  /*
   * exception 255 indicates a software trap
   * inserted in place of code ... so back up
   * PC by one instruction, since this instruction
   * will later be replaced by its original one!
   */
  if (exceptionVector == 0xff
      || exceptionVector == 0x20)
    registers[PC] -= 2;

  /*
   * Do the thangs needed to undo
   * any stepping we may have done!
   */
  undoSStep ();

  while (1)
    {
      remcomOutBuffer[0] = 0;
      getpacket (remcomInBuffer);

      switch (remcomInBuffer[0])
	{
	case '?':
	  remcomOutBuffer[0] = 'S';
	  remcomOutBuffer[1] = highhex (sigval);
	  remcomOutBuffer[2] = lowhex (sigval);
	  remcomOutBuffer[3] = 0;
	  break;
	case 'd':
	  remote_debug = !(remote_debug);	/* toggle debug flag */
	  break;
	case 'g':		/* return the value of the CPU registers */
	  mem2hex ((char *) registers, remcomOutBuffer, NUMREGBYTES);
	  break;
	case 'G':		/* set the value of the CPU registers - return OK */
	  hex2mem (&remcomInBuffer[1], (char *) registers, NUMREGBYTES);
	  strcpy (remcomOutBuffer, "OK");
	  break;

	  /* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
	case 'm':
	  if (setjmp (remcomEnv) == 0)
	    {
	      dofault = 0;
	      /* TRY, TO READ %x,%x.  IF SUCCEED, SET PTR = 0 */
	      ptr = &remcomInBuffer[1];
	      if (hexToInt (&ptr, &addr))
		if (*(ptr++) == ',')
		  if (hexToInt (&ptr, &length))
		    {
		      ptr = 0;
		      mem2hex ((char *) addr, remcomOutBuffer, length);
		    }
	      if (ptr)
		strcpy (remcomOutBuffer, "E01");
	    }
	  else
	    strcpy (remcomOutBuffer, "E03");

	  /* restore handler for bus error */
	  dofault = 1;
	  break;

	  /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
	case 'M':
	  if (setjmp (remcomEnv) == 0)
	    {
	      dofault = 0;

	      /* TRY, TO READ '%x,%x:'.  IF SUCCEED, SET PTR = 0 */
	      ptr = &remcomInBuffer[1];
	      if (hexToInt (&ptr, &addr))
		if (*(ptr++) == ',')
		  if (hexToInt (&ptr, &length))
		    if (*(ptr++) == ':')
		      {
			hex2mem (ptr, (char *) addr, length);
			ptr = 0;
			strcpy (remcomOutBuffer, "OK");
		      }
	      if (ptr)
		strcpy (remcomOutBuffer, "E02");
	    }
	  else
	    strcpy (remcomOutBuffer, "E03");

	  /* restore handler for bus error */
	  dofault = 1;
	  break;

	  /* cAA..AA    Continue at address AA..AA(optional) */
	  /* sAA..AA   Step one instruction from AA..AA(optional) */
	case 'c':
	case 's':
	  {
	    /* tRY, to read optional parameter, pc unchanged if no parm */
	    ptr = &remcomInBuffer[1];
	    if (hexToInt (&ptr, &addr))
	      registers[PC] = addr;

	    if (remcomInBuffer[0] == 's')
	      doSStep ();
	  }
	  return;
	  break;

	  /* kill the program */
	case 'k':		/* do nothing */
	  break;
	}			/* switch */

      /* reply to the request */
      putpacket (remcomOutBuffer);
    }
}


#define GDBCOOKIE 0x5ac 
static int ingdbmode;
/* We've had an exception - choose to go into the monitor or
   the gdb stub */
void handle_exception(int exceptionVector)
{
#ifdef MONITOR
    if (ingdbmode != GDBCOOKIE) {
		monitor_handle_exception (exceptionVector);
		return;
	}
#endif
	if (exceptionVector == BIOS_CALL_TRAP) {
		handle_bios_call ();
		return;
	}

	gdb_handle_exception (exceptionVector);
}

void
gdb_mode()
{
  ingdbmode = GDBCOOKIE;
  breakpoint();
}
/* This function will generate a breakpoint exception.  It is used at the
   beginning of a program to sync up with a debugger and can be used
   otherwise as a quick means to stop program execution and "break" into
   the debugger. */

void
breakpoint (void)
{
      BREAKPOINT ();
}


/*
 *  ٥١쥸VBRˤ
 */
static __inline__ void
set_vbr(void *vbr)
{
	__asm__ volatile("ldc  %0, vbr" : : "r"(vbr) );
}


/*  ߴطν  */
void
init_int(void)
{
	/*  ٥ơ֥ν  */

	/*
	 * XXX: ٥ơ֥ƬĴ٤Ū
	 * ʤɬפǤ롣
	 */
	memcpy((void *)vectable, (void *)0, vector_size*4);	//0Ϥ饳ԡ
	
	/*  ٥١쥸VBRˤ  */
	set_vbr((void *)vectable);

	/*
	 * ߥȥν
	 */
	*IPRA = 0x0000;
	*IPRB = 0x0000;
	*IPRC = 0x0000;
	*IPRD = 0x0000;
	*IPRE = 0x0000;
	*IPRG = 0x0000;
	*IPRH = 0x0000;
}

/*  ꥻåȸ塢bssdata  */
/*  Ƥ餳Ǥ  */
// stub_start̾Ƥ
void
INIT (void)
{
  /* First turn on the ram */
  // stub_start

  init_int();	/*  ߴطν  */
  init_serial();

#ifdef MONITOR
  reset_hook ();
#endif


  in_nmi = 0;
  dofault = 1;
  stepped = 0;

  stub_sp = stub_stack + stub_stack_size;

  breakpoint ();  // trapa #0x20

  while (1)
    ;
}


static void sr()
{


  /* Calling Reset does the same as pressing the button */
  asm (".global _Reset
        .global _WarmReset
_Reset:
_WarmReset:
         mov.l L_sp,r15
         bra   _INIT
         nop
         .align 2
L_sp:    .long _init_stack + " TO_STRING(init_stack_size) );

  asm(" .align 2
saveRegisters:
	mov.l	@(L_reg, pc), r0
	mov.l	@r15+, r1				! pop R0
	mov.l	r2, @(0x08, r0)				! save R2
	mov.l	r1, @r0					! save R0
	mov.l	@r15+, r1				! pop R1
	mov.l	r3, @(0x0c, r0)				! save R3
	mov.l	r1, @(0x04, r0)				! save R1
	mov.l	r4, @(0x10, r0)				! save R4
	mov.l	r5, @(0x14, r0)				! save R5
	mov.l	r6, @(0x18, r0)				! save R6
	mov.l	r7, @(0x1c, r0)				! save R7
	mov.l	r8, @(0x20, r0)				! save R8
	mov.l	r9, @(0x24, r0)				! save R9
	mov.l	r10, @(0x28, r0)			! save R10
	mov.l	r11, @(0x2c, r0)			! save R11
	mov.l	r12, @(0x30, r0)			! save R12
	mov.l	r13, @(0x34, r0)			! save R13
	mov.l	r14, @(0x38, r0)			! save R14
	mov.l	@r15+, r4				! save arg to handleException
	add	#8, r15					! hide PC/SR values on stack
	mov.l	r15, @(0x3c, r0)			! save R15
	add	#-8, r15				! save still needs old SP value
	add	#92, r0					! readjust register pointer
	mov	r15, r2
	add	#4, r2
	mov.l	@r2, r2					! R2 has SR
	mov.l	@r15, r1				! R1 has PC
	mov.l	r2, @-r0				! save SR
	sts.l	macl, @-r0				! save MACL
	sts.l	mach, @-r0				! save MACH
	stc.l	vbr, @-r0				! save VBR
	stc.l	gbr, @-r0				! save GBR
	sts.l	pr, @-r0				! save PR
	mov.l	@(L_stubstack, pc), r2
	mov.l	@(L_hdl_except, pc), r3
	mov.l	@r2, r15
	jsr	@r3
	mov.l	r1, @-r0				! save PC
	mov.l	@(L_stubstack, pc), r0
	mov.l	@(L_reg, pc), r1
	bra	restoreRegisters
	mov.l	r15, @r0				! save __stub_stack
	
	.align 2
L_reg:
	.long	_registers
L_stubstack:
	.long	_stub_sp
L_hdl_except:
	.long	_handle_exception");

}

static void rr()
{
asm("
	.align 2	
        .global _resume
_resume:
	mov	r4,r1
restoreRegisters:
	add	#8, r1						! skip to R2
	mov.l	@r1+, r2					! restore R2
	mov.l	@r1+, r3					! restore R3
	mov.l	@r1+, r4					! restore R4
	mov.l	@r1+, r5					! restore R5
	mov.l	@r1+, r6					! restore R6
	mov.l	@r1+, r7					! restore R7
	mov.l	@r1+, r8					! restore R8
	mov.l	@r1+, r9					! restore R9
	mov.l	@r1+, r10					! restore R10
	mov.l	@r1+, r11					! restore R11
	mov.l	@r1+, r12					! restore R12
	mov.l	@r1+, r13					! restore R13
	mov.l	@r1+, r14					! restore R14
	mov.l	@r1+, r15					! restore programs stack
	mov.l	@r1+, r0
	add	#-8, r15					! uncover PC/SR on stack 
	mov.l	r0, @r15					! restore PC onto stack
	lds.l	@r1+, pr					! restore PR
	ldc.l	@r1+, gbr					! restore GBR		
	ldc.l	@r1+, vbr					! restore VBR
	lds.l	@r1+, mach					! restore MACH
	lds.l	@r1+, macl					! restore MACL
	mov.l	@r1, r0	
	add	#-88, r1					! readjust reg pointer to R1
	mov.l	r0, @(4, r15)					! restore SR onto stack+4
	mov.l	r2, @-r15
	mov.l	L_in_nmi, r0
	mov		#0, r2
	mov.b	r2, @r0
	mov.l	@r15+, r2
	mov.l	@r1+, r0					! restore R0
	rte
	mov.l	@r1, r1						! restore R1

");
}

static __inline__ void code_for_catch_exception(int n) 
{
  asm("		.global	_catch_exception_%O0" : : "i" (n) 				); 
  asm("	_catch_exception_%O0:" :: "i" (n)      						);

  asm("		add	#-4, r15 				! reserve spot on stack ");
  asm("		mov.l	r1, @-r15				! push R1		");

  if (n == NMI_VEC) 
    {
      /* Special case for NMI - make sure that they don't nest */
      asm("	mov.l	r0, @-r15					! push R0");
      asm("	mov.l	L_in_nmi, r0");
      asm("	tas.b	@r0						! Fend off against addtnl NMIs");
      asm("	bt		noNMI");
      asm("	mov.l	@r15+, r0");
      asm("	mov.l	@r15+, r1");
      asm("	add		#4, r15");
      asm("	rte");
      asm("	nop");
      asm(".align 2");
      asm("L_in_nmi: .long	_in_nmi");
      asm("noNMI:");
    }
  else
    {
      if (n == CPU_BUS_ERROR_VEC)
	{
	  /* Exception 9 (bus errors) are disasbleable - so that you
	     can probe memory and get zero instead of a fault.
	     Because the vector table may be in ROM we don't revector
	     the interrupt like all the other stubs, we check in here
	     */
	  asm("mov.l	L_dofault,r1");
	  asm("mov.l	@r1,r1");
	  asm("tst	r1,r1");
	  asm("bf	faultaway");
	  asm("bsr	_handle_buserror");
	  asm(".align	2");
	  asm("L_dofault: .long _dofault");
	  asm("faultaway:");
	}
      asm("		mov	#15<<4, r1							");
      asm("		ldc	r1, sr					! disable interrupts	");
      asm("		mov.l	r0, @-r15				! push R0		");
    }

  /* Prepare for saving context, we've already pushed r0 and r1, stick exception number
     into the frame */
  asm("		mov	r15, r0								");
  asm("		add	#8, r0								");
  asm("		mov	%0,r1" :: "i" (n)        					);
  asm("		extu.b  r1,r1								");
  asm("		bra	saveRegisters				! save register values	");
  asm("		mov.l	r1, @r0					! save exception # 	");
}


static  void
exceptions()
{
  code_for_catch_exception (CPU_BUS_ERROR_VEC);
  code_for_catch_exception (DMA_BUS_ERROR_VEC);
  code_for_catch_exception (INVALID_INSN_VEC);
  code_for_catch_exception (INVALID_SLOT_VEC);
  code_for_catch_exception (NMI_VEC);
  code_for_catch_exception (TRAP_VEC);
  code_for_catch_exception (BIOS_CALL_TRAP);

/*  remote_break  
 *	^C(\003)ߤˤ⡼ȥ֥졼
 *	ƤФȤˤϳ߶ػߤˤʤäƤIPM=0xf
 *	Τᡢ2ʸܰʹߤϼߤˤʤʤ
 *	stub˼ߤϼդʤ
 *	ݡ󥰤ˤ
 *	1ʸɤ߹ϥ桼֥졼ƱԤ
 */
  asm(" .global _remote_break
        .align 2
_remote_break:
	mov.l	r0, @-r15	! store r0-r7, pr
	mov.l	r1, @-r15
	mov.l	r2, @-r15
	mov.l	r3, @-r15
	mov.l	r4, @-r15
	mov.l	r5, @-r15
	mov.l	r6, @-r15
	mov.l	r7, @-r15
	sts.l	pr, @-r15
				! depend on serial device
	mov.l	L_handle_remote_break, r0
	jsr	@r0
	nop
	
	lds.l	@r15+, pr	! load pr, r0-r7
	mov.l	@r15+, r7
	mov.l	@r15+, r6
	mov.l	@r15+, r5
	mov.l	@r15+, r4
	mov.l	@r15+, r3
	mov.l	@r15+, r2
	mov.l	@r15+, r1
	mov.l	@r15+, r0
	");

/*  1ʸɤ߹ϥ桼֥졼ƱԤ  */
  code_for_catch_exception (USER_VEC);
  asm("	.align 2
L_handle_remote_break:
	.long	_handle_remote_break");

}

/* 
Ѥɤ٥ֹ椫å
   
   n: å٥ֹ
   
   OK
   NG
 */
static int
is_vector_ok (int n)
{
	int i, size;
	
	if (n <= 4) return 0;		/*  ꥻåȥ٥  */
	if (n >= 196) return 0;		/*  196ʹߤϤ٤ͽѤ  */
					/*  ¾ͽѤֹ  */
	size = ARRAY_SIZE(reservation_list);
	for (i=0; i<size; i++) {
		if (n==reservation_list[i]) return 0;
	}
	return 1;
}

/* 
ߥϥɥϿ
   
   n: ٥ֹ
   func: ؿؤΥݥ
   
   OK
   ~NG
 */
static int
set_handler(int n, Function func)
{
	if (is_vector_ok(n)) {
		vectable[n]=func;
		return 0;
	}
	return ~0;
}

/* 
BIOS
   R0: Function Number
   R4-R7: Input Arguments
   R0: Return value
 */
static void
handle_bios_call (void)
{
  unsigned int func = registers[R0];
  unsigned int arg0 = registers[R4];
  unsigned int arg1 = registers[R5];
//  unsigned int arg2 = registers[R6];
//  unsigned int arg3 = registers[R7];
  unsigned int ret;

  switch (func)
    {
    case 0:
        /*
         *  gdbconsoleؤν
         *  break뤳Ȥʤgcb˽Ϥϡ
         *  /gdb-4.18/gdb/remote.c
         *  fputs_filtered (tb, gdb_stdout);
         *  +gdb_flush(gdb_stdout);
         *  ɲä롣
         */        
        if (ingdbmode)
        {
            remcomOutBuffer[0] = 'O';
            remcomOutBuffer[1] = highhex(arg0);
            remcomOutBuffer[2] = lowhex (arg0);
            remcomOutBuffer[3] = 0;
            putpacket (remcomOutBuffer);
        }
        else
        {
            int ch;
            char *p = (char *)arg0;
            
            while ((ch = *p++))
                putDebugChar (ch);
        }
        ret = 0;
        break;

    case 1:
      /* Second strage access */
      /* Not implemented yet */
      ret = ~0;
      break;
    case 8:
        /*
         * vectable˥ϥɥϿ롣
         * arg0Υ٥ֹ桢arg1˥ϥɥƬɥ쥹ꤹ
         */
        ret = set_handler(arg0, (Function)arg1);        
        break;
    case 255:
      /* Detach gdb mode */
      ingdbmode = 0;
      putpacket ("W00");
      getDebugChar ();
      ret = 0;
      break;

    default:
      /* Do nothing */
      ret = ~0;
      break;
    }

  registers[R0] = ret;
}


/* ϡɥ */
void
hardware_init_hook (void)
{
	*BCR1 = 0x202f;
	*BCR2 = 0x5500;
	*WCR1 = 0x0010;
	*PACRH = 0x5000;
	*PACRL1 = 0x1540;
	*PACRL2 = 0x0140;
	*PBCR1 = 0x0002;
	*PBCR2 = 0xa005;
	*PCCR = 0xfffc;
	*PDCRH1 = 0x5555;
	*PDCRH2 = 0x5555;
	*PDCRL1 = 0xffff;
	*PDCRL2 = 0x0000;
	*PECRL1 = 0x0000;
	*PECRL2 = 0x0000;
	*PAIORL = 0x87e7;
	*PBIOR = 0x023c;
}

